#!/usr/bin/env python


__author__ = "Ivan Prieto Barreiro"
__email__ = "ivan.prieto.barreiro@cern.ch"
__copyright__ = "(C) 2011 European Organization for Nuclear Research, Geneva, Switzerland"
__version__ = "1.3"
__date__ = "$Date: 2012-02-03 20:54:05 +0100 (Fri, 03 Feb 2012) $"
__description__ = "Plugin for FlexExtractor to handle XML files using XPATH"
__status__ = "Production" 




# =============================================
#  FUNCTION RESPONSIBLE FOR READING DATA FROM A CERTAIN
#  INPUT. THIS FUNCTION IS CALLED BY FlexExtractor.
# =============================================
def getData(pFile, pSourceList):
    """
    This function reads data from an input source specified in the parameter "pFile". The structure
    defined in "pSourceList" is used to know what data to read from the input source. This function
    returns a tuple with four elements: code, header, data and message.

    Parameter(s):
       pFile - string containing the name of the input source to be read.     
       pSourceList - list containing the input structure that was specified in the XML config.
    
    Return(s):
       int - 0 (the input source was successfully read) or -1 (error found when reading the input source).
       list - contains a list with the headers of the input source (this list should be empty when the "autoselect"
              is False and filled-up when "autoselect" is True).                    
       list - contains the data that was read from the input source (this list will be empty if some error was found).
       list - contains a list of messages that were generated during the reading of the input source.             
    """
    
    try:
        import os
    except:
        return -1, [], [], ["Error when importing module 'os'."]        
    
    try:
        from xml.dom import minidom
        from xml.parsers.expat import ExpatError        
    except:
        return -1, [], [], ["Error when importing module 'minidom'."]
            
    try:
        from xpath.expr import xpath
    except:
        return -1, [], [], ["Error when importing module 'xpath'."]

    if not os.path.isfile(pFile):
        return -1, [], [], ["The input file '%s' does not exists." % pFile]
    
    try:
        parser = minidom.parse(pFile)
        messageList = []        
        result = []           
        for source in pSourceList:
            if source["parameter"] is None:                       
                nodeset = xpath.find("/", parser)
            else:
                nodeset = xpath.find(source["parameter"], parser)
            data = []
            for node in nodeset:
                tmpData = []
                for expression in source["data"]:
                    if expression["select"] is None:
                        value = None
                    else:                    
                        value = xpath.findvalue(expression["select"], node)
                        if value is None:
                            message = "The input file '%s' does not have a select '%s' in source '%s' or it is invalid." % (pFile, expression["select"], source["name"])
                            if message not in messageList:
                                messageList.append(message)                        
                    tmpData.append(value)
                data.append(tmpData)
            result.append(data)        
        return 0, [], result, messageList     
    except MemoryError:
        return -1, [], [], ["Error when reading input file '%s' since there is not enough memory available." % pFile]
    except ExpatError, expatError:
        return -1, [], [], ["Error when parsing input file '%s' in line '%s' at column '%s'." % (pFile, expatError.lineno, expatError.offset + 1)]
    except Exception, e:
        print e
        return -1, [], [], ["Error when reading input file '%s'." % pFile]        





# =============================================
#  FUNCTION RESPONSIBLE FOR WRITING DATA TO A CERTAIN
#  OUTPUT. THIS FUNCTION IS CALLED BY FlexExtractor.
# =============================================
def setData(pFile, pParameter, pDataList, pAppendFlag):
    """
    This function writes the data "pDataList" into a certain output specified in the parameter "pFile"
    and uses the parameter "pParameter" to provide additional information on how to do it. If "pAppendFlag"
    is True then the data will be appended in the end of the output, otherwise (if False) the previous
    data will be erased before writing the new data. This function returns a tuple with two elements:
    code and message.

    Parameter(s):
       pFile - string containing the name of the output where to write the data.    
       pParameter - string containing the parameter that was specified in the XML config.    
       pDataList - list containing the data to be written.
       pAppendFlag - boolean that specifies if the output is to be overwritten (True) or appended (False) with the new data.
    
    Return(s):
       int - 0 (the data was successfully written) or -1 (error found when writing the data).             
       list - contains a list of messages that were generated during the writing of the data.      
    """
    
    return 0, []
    

