package org.xdv.clx.builder;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xdv.clx.base.ClxConstants;
import org.xdv.clx.expr.ClxAnd;
import org.xdv.clx.expr.ClxEqual;
import org.xdv.clx.expr.ClxExists;
import org.xdv.clx.expr.ClxForAll;
import org.xdv.clx.expr.ClxFormula;
import org.xdv.clx.expr.ClxFormulaTrue;
import org.xdv.clx.expr.ClxGreater;
import org.xdv.clx.expr.ClxGreaterOrEqual;
import org.xdv.clx.expr.ClxHeader;
import org.xdv.clx.expr.ClxIff;
import org.xdv.clx.expr.ClxImplies;
import org.xdv.clx.expr.ClxKey;
import org.xdv.clx.expr.ClxLess;
import org.xdv.clx.expr.ClxLessOrEqual;
import org.xdv.clx.expr.ClxMacro;
import org.xdv.clx.expr.ClxMacroDefinitions;
import org.xdv.clx.expr.ClxNot;
import org.xdv.clx.expr.ClxNotEqual;
import org.xdv.clx.expr.ClxOr;
import org.xdv.clx.expr.ClxRule;
import org.xdv.clx.expr.ClxRules;
import org.xdv.clx.expr.ClxSame;
import org.xdv.clx.expr.ClxVariable;
import org.xdv.clx.extension.ClxExtErrorMessage;
import org.xdv.clx.extension.ClxExtErrorMessageText;
import org.xdv.clx.extension.ClxExtErrorMessageXPath;
import org.xdv.clx.extension.ClxExtErrorReport;
import org.xdv.clx.extension.ClxExtMacroCall;
import org.xdv.clx.extension.ClxExtMacroCallPostPrep;
import org.xdv.clx.operator.ClxOperatorCall;
import org.xdv.clx.operator.ClxOperatorCallPostPrep;
import org.xdv.clx.operator.ClxOperatorParam;
import org.xdv.common.UtilDOM;
import org.xdv.common.XdvValidationError;
import org.xdv.xpath.XPathEngine;
import org.xdv.xpath.XPathException;
import org.xdv.xpath.XPathExpression;
import org.xdv.xpath.XPathUtil;

/**
 * CLX builder to create CLX element from W3C DOM elements.
 * <p>
 * This class is dedicated to construct/build CLX elements (like for instance 
 * {@link ClxRules}) from W3C DOM elements ({@link Element}).
 * <br>
 * Builder process source document tree and it fails only in case of fatal error, like:
 * <ul>
 *  <li>missing attribute or element,
 *  <li>error in XPath query,
 *  <li>invalid number of elements (for instance binary operator requires exacly two direct children).
 * </ul>  
 * Builder doesn't check other types of XML Schema constraints (like for instance XML data-types,
 * elements order, etc). 
 * </p>
 * <p>
 * Builder doesn't operate on W3C DOM source directly but only through so called <i>helper</i>
 * ({@link ClxDOMHelper}). Helper may be namespace aware or not. 
 * DOM elements may come from 3 different namespaces:
 * <ol>
 *  <li>CLX namespace - global CLX hierarchy defined in 
 *      <a href="http://www.clixml.org">CLiXML.org</a> ({@link ClxConstants#NS_CLIXML}),  
 *  <li>macroCall namespace - extension (specific to this implementation) to express macro call
 *      ({@link ClxConstants#NS_EXT_MACRO}),
 *  <li>error-report namespace - extnesion (specific to this implementation) to define
 *      error-report ({@link ClxConstants#NS_EXT_REPORT}).
 * </ol>       
 * </p>
 * <p>
 * This implementation produces {@link ClxExtMacroCallPostPrep} as a result of macroCall and
 * {@link ClxOperatorCallPostPrep} as a result of operator call. 
 * It means that one should <i>post prepare</i> all those calls before executing any formulas
 * created by this builder.
 * </p>
 * <p>
 * It is not a thread safe implementation.
 * </p>
 *
 * @author Tomasz Kokoszka
 */
public class ClxDOMBuilder {
    
    /** XPath engine to create XPath expressions required by some CLX elements */
    private final XPathEngine xpathEngine;
    
    /** General helper to operate on CLX hierarchy */
    private final ClxDOMHelper helper;   
    
    /** 
     * Does builder accept macroCall elements that come from <i>empty</i> namespace?
     * If <code>false</code>, every macroCall element must be in proper namespace.
     * <br>Note: to work with namespaces you have to use namespace aware parser. 
     */
    private boolean macroNoNS = true;
    
    /** 
     * Does builder accepts report elements that come from <i>empty</i> namespace?
     * If <code>false</code>, every report element must be in proper namespace.
     * <br>Note: to work with namespaces you have to use namespace aware parser.
     */
    private boolean reportNoNS = true;
    
    /**
     * Does builder should skip rule header?
     * Value <b><code>true</code></b> means to skip rule header - it will be always set to 
     * <code>null</code>. 
     * <b><code>False</code></b> means to load rule header and attach it to the rule.
     */
    private boolean skipHeader = true;
    
    
    /**
     * List of all created macro calls - list of {@link ClxExtMacroCall}.
     * Attribute is used for internal compuatations.
     */
    private List createdMacroCalls = null;
    
    /**
     * List of all created operator calls - list of {@link ClxOperatorCall}.
     * Attribute is used for internal compuatations.
     */
    private List createdOperatorCalls = null;
    
    /**
     * Available XPath variables - set of {@link String}.
     * To check if XPath expression uses only variable that have been declared before.
     * Attribute is used for internal compuatations.
     */
    private Set availableVars = null;
    
    /**
     * Available XPath variables (for error message evaluation) - set of {@link String}.
     * To check if XPath expression in error message uses only variable that have been
     * declared before.
     */
    private Set availableVarsError = null;
    
    
    
    /**
     * Create a builder.
     * @param engine XPath engine to create XPath expression required by some CLX elements
     * @param helper interface to operate on CLX hierarchy
     */
    public ClxDOMBuilder(XPathEngine engine, ClxDOMHelper helper) {
        this.xpathEngine = engine;
        this.helper = helper;
    }
    
    /**
     * Set builder behavoir for error report definition.
     * @param allowNoNS should builder accept report definition elements that come from 
     *        <i>empty</i> namespace?
     */
    public void setAllowReportNoNS(boolean allowNoNS) {
        this.reportNoNS = allowNoNS;
    }
    
    /**
     * Set builder behavoir for macroCall.
     * @param allowNoNS should builder accept macroCall that come from 
     *        <i>empty</i> namespace?
     */
    public void setAllowMacroNoNS(boolean allowNoNS) {
        this.macroNoNS = allowNoNS;
    }
    
    /**
     * Set builder behaviur for rule header.
     * Value <b><code>true</code></b> means to skip rule header - it will be always set to 
     * <code>null</code>. 
     * <b><code>False</code></b> means to load rule header and attach it to the rule.
     */
    public void setSkipHeader(boolean skipHeader) {
        this.skipHeader = skipHeader;        
    }
    
    /**
     * Creates XPath expression using {@link #xpathEngine}.
     * Checks if all variables from expression are in {@link #availableVars}. 
     * @param xpath xpath as a text, for instance <code>/data/field[@name='xx']</code> 
     * @param errorElement element definition - used only to generate exception message when
     *        XPath expression creation fails
     * @return xpath expression; !=null
     * @throws ClxBuilderException error while creating XPath expression - syntax error, or 
     *         missing variable
     */
    private XPathExpression createXPath(String xpath, Element errorElement) 
    throws ClxBuilderException {
        return createXPath(xpath, errorElement, this.availableVars);
    }
        
    /**
     * Creates XPath expression using {@link #xpathEngine}.
     * @param xpath xpath as a text, for instance <code>/data/field[@name='xx']</code> 
     * @param errorElement element definition - used only to generate exception message when
     *        XPath expression creation fails
     * @param variables names of declared (available variables), set of {@link String}
     * @return xpath expression; !=null
     * @throws ClxBuilderException error while creating XPath expression - syntax error, or 
     *         missing variable
     */
    private XPathExpression createXPath(String xpath, Element errorElement, Set variables) 
    throws ClxBuilderException {
        XPathExpression expression = null;
        try {
            expression = xpathEngine.createExpression(xpath);
        } catch (XPathException e) {
            throw new ClxBuilderException("Unable to compile xpath " + xpath, 
                    getErrorPath(errorElement), e);
        }
        
        //check if all required variables have been declared
        Set requiredVars = XPathUtil.extractVariableNames(xpath);
        if (! requiredVars.isEmpty()) {
            requiredVars.removeAll(variables);
            if (! requiredVars.isEmpty()) {
                throw new ClxBuilderException("Variables " + requiredVars.toString() + 
                        " are used in xpath but have not been declared, xpath: " + xpath, 
                        getErrorPath(errorElement));
            }
        }
        
        return expression;
    }
    
    /**
     * Declares usage of a variable.
     * Adds variable name to {@link #availableVars} and {@link #availableVarsError}.
     * @param varName variable name
     * @param errorElement node that will be reported in case of any error
     * @throws ClxBuilderException variable has been already declared
     */
    private void declareVariable(String varName, Element errorElement)
    throws ClxBuilderException {
        if (! this.availableVars.add(varName)) {
            throw new ClxBuilderException("Variable " + varName + " already in use", 
                    getErrorPath(errorElement));
        }
        this.availableVarsError.add(varName);
    }
    
    /**
     * Returns values of all elements from source list.
     * For each element from source list: take its value, format it and put into the result list.
     * It uses:
     * <ul>
     *  <li>{@link UtilDOM#getElementValue(Element)} - to take element value
     *  <li>{@link UtilDOM#formatWhitespaces(String, boolean, boolean)} - to format value
     * </ul>
     * @param elements list of {@link Element}
     * @return list of {@link String}; !=null
     */
    private List concatElemsValue(List elements) {
        final List result = new ArrayList(elements.size());
        for(Iterator i = elements.iterator(); i.hasNext(); ) {
            final Element e = (Element) i.next();
            result.add(UtilDOM.formatWhitespaces(UtilDOM.getElementValue(e), true, true));
        }
        return result;
    }
    
    /**
     * Creates path to given element - it is used to generate path for {@link ClxBuilderException}. 
     * @param elem XML element
     * @return path, for instance <code>/data/field</code>; !=null.
     */
    private String getErrorPath(Node elem) {
        return UtilDOM.getSimplePath(elem);
    }
    
    /**
     * Get children that are (potential) formulas.
     * Child is an potencial formula if it comes from {@link #helper} namespace  
     * or {@link ClxConstants#NS_EXT_MACRO} namespace.
     * @param root XML element
     * @return list of {@link Element}; !=null
     */
    private List getChildrenFormulas(Element root) 
    throws ClxBuilderException {
        return helper.getChildrenWithNS(root, ClxConstants.NS_EXT_MACRO, macroNoNS);        
    }        
    
    /**
     * Creates a macro definition object.
     * <br>
     * Remember to <i>post process</i> all macro and operator calls before using created 
     * macros. 
     * @param macroDefElem source - W3C DOM element - {@link ClxConstants#TAG_MACRODEFINITIONS} 
     * @return macro definitions; !=null
     * @throws ClxBuilderException error in source
     */
    public ClxMacroDefinitions createMacroDefinitions(Element macroDefElem) 
    throws ClxBuilderException {
        this.createdMacroCalls = new ArrayList();
        this.createdOperatorCalls = new ArrayList();
        this.availableVars = new HashSet();
        this.availableVarsError = new HashSet();
        
        final ClxHeader header = 
            createHeaderNull(helper.getFirstOne(macroDefElem, ClxConstants.TAG_HEADER));
        
        final ClxMacroDefinitions result = new ClxMacroDefinitions(header);
    
        //parse each macro
        List macroElems = helper.getChildren(macroDefElem, ClxConstants.TAG_MACRO);
        for(Iterator i = macroElems.iterator(); i.hasNext(); ) {
            final Element macroElem = (Element) i.next();
            final ClxMacro macro = createMacro(macroElem);
            
            if (result.addMacro(macro) != null) {
                throw new ClxBuilderException("Macro " + macro.getName() + " already defined", 
                        getErrorPath(macroElem));
            }
            
        }
        macroElems = null;
        
        if (result.getMacroNames().isEmpty()) {
            throw new ClxBuilderException("At least one macro definition required",
                    getErrorPath(macroDefElem));
        }
        
        result.setMacroCalls(this.createdMacroCalls);
        result.setOperatorCalls(this.createdOperatorCalls);
        this.createdMacroCalls = null;
        this.createdOperatorCalls = null;
        this.availableVars = null;
        this.availableVarsError = null;
        
        return result;        
    }
    
    /**
     * Creates a rule set object.
     * <br>
     * Remember to <i>post process</i> all macro and operator calls before using created 
     * rules. 
     * @param rulesElem source - W3C DOM element - {@link ClxConstants#TAG_RULES} 
     * @return rules definitions; !=null
     * @throws ClxBuilderException error in source
     */
    public ClxRules createRules(Element rulesElem)
    throws ClxBuilderException {
        this.createdMacroCalls = new ArrayList();
        this.createdOperatorCalls = new ArrayList();
        this.availableVars = new HashSet();
        this.availableVarsError = new HashSet();        
        
        final List macroHrefs = new ArrayList();
        final List variables = new ArrayList();
        final List keys = new ArrayList();
        final List rules = new ArrayList();
        
        List children = helper.getChildren(rulesElem);
        for(Iterator i = children.iterator(); i.hasNext(); ) {
            final Element elem = (Element) i.next();
            final String elemName = helper.getNodeName(elem);
            
            if (ClxConstants.TAG_MACROS.equals(elemName)) {
                macroHrefs.add(createMacrosHref(elem));                
            } else if (ClxConstants.TAG_VARIABLE.equals(elemName)) {                
                variables.add(createVariable(elem));
            } else if (ClxConstants.TAG_KEY.equals(elemName)) {
                keys.add(createKey(elem));
            } else if (ClxConstants.TAG_RULE.equals(elemName)) {
                ClxRule rule = createRule(elem);
                rules.add(rule);
            } else {
                throw new ClxBuilderException("Element " + elemName + " not expected in rule set", 
                        getErrorPath(elem));
            }
        }
        children = null; //I don't need this value in further compuations
        
        final ClxRules result = new ClxRules(macroHrefs, variables, keys);
        
        //set rules
        if (rules.isEmpty()) {
            throw new ClxBuilderException("At least on rule is required", getErrorPath(rulesElem));
        }
        ClxRule duplicate = result.addRuleAll(rules);
        if (duplicate != null) {
            throw new ClxBuilderException("Rule " + duplicate.getId() + " already defined", 
                    getErrorPath(rulesElem));
        }
                        
        result.setMacroCalls(this.createdMacroCalls);
        result.setOperatorCalls(this.createdOperatorCalls);
        this.createdMacroCalls = null;
        this.createdOperatorCalls = null;
        this.availableVars = null;
        this.availableVarsError = null;
        
        return result;
    }
    
    /**
     * Creates a single macro definition object.
     * @param macroElem source - W3C DOM element - {@link ClxConstants#TAG_MACRO} 
     * @return macro definitions; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxMacro createMacro(Element macroElem) 
    throws ClxBuilderException {        
        final String name = helper.getAttrValueNoEmpty(macroElem, ClxConstants.ATTR_NAME);
        final ClxHeader header = createHeaderNull(
                helper.getFirstOne(macroElem, ClxConstants.TAG_HEADER));        
        final Set params = new HashSet();
        
        //for each param:
        List paramElems = helper.getChildren(macroElem, ClxConstants.TAG_PARAM);        
        for(Iterator i = paramElems.iterator(); i.hasNext(); ) {
            final Element paramElem = (Element) i.next();
            final String paramName = helper.getAttrValueNoEmpty(paramElem, ClxConstants.ATTR_NAME);
            
            declareVariable(paramName, paramElem);
            params.add(paramName);
        }
        paramElems = null;
                
        //build output
        Element outputElem = helper.getFirstOne(macroElem, ClxConstants.TAG_OUTPUT);
        if (outputElem == null) {
            throw new ClxBuilderException("Element " + ClxConstants.TAG_OUTPUT + " is missing " +
                    "for macro " + name, getErrorPath(macroElem));
        }
        List formulaElems = getChildrenFormulas(outputElem);
        if (formulaElems.size() != 1) {
            throw new ClxBuilderException("Exacly one formula definition expected for " +
                    "macro " + name, getErrorPath(outputElem));
        }
        final ClxFormula output = createFormula((Element) formulaElems.get(0));
        
        this.availableVars.removeAll(params);
                        
        return new ClxMacro(name, params, output, header); 
    }               
    
    /**
     * Creates URL ref to a macros definitions.
     * @param macrosElem source - W3C DOM element - {@link ClxConstants#TAG_MACROS} 
     * @return URL to macros definitions; !=null
     * @throws ClxBuilderException error in source
     */
    private String createMacrosHref(Element macrosElem)
    throws ClxBuilderException {
        return helper.getAttrValueNoEmpty(macrosElem, ClxConstants.ATTR_HREF);        
    }
    
    /**
     * Creates a rules global variable.
     * @param variableElem source - W3C DOM element - {@link ClxConstants#TAG_VARIABLE} 
     * @return variable definition; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxVariable createVariable(Element variableElem)
    throws ClxBuilderException {
        String id = helper.getAttrValueNoEmpty(variableElem, ClxConstants.ATTR_ID);
        String xpath = helper.getAttrValueNoEmpty(variableElem, ClxConstants.ATTR_XPATH);        
        
        declareVariable(id, variableElem);
        
        return new ClxVariable(id, createXPath(xpath, variableElem));
    }
    
    /**
     * Creates a key binding.
     * @param keyElem source - W3C DOM element - {@link ClxConstants#TAG_KEY} 
     * @return key definition; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxKey createKey(Element keyElem) 
    throws ClxBuilderException {
        String name = helper.getAttrValueNoEmpty(keyElem, ClxConstants.ATTR_NAME);
        String match = helper.getAttrValueNoEmpty(keyElem, ClxConstants.ATTR_MATCH);
        String use = helper.getAttrValueNoEmpty(keyElem, ClxConstants.ATTR_USE);
        
        return new ClxKey(name, createXPath(match, keyElem), createXPath(use, keyElem));
    }
    
    /**
     * Creates a rule.
     * @param ruleElem source - W3C DOM element - {@link ClxConstants#TAG_RULE} 
     * @return rule definition; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxRule createRule(Element ruleElem) 
    throws ClxBuilderException {
        final Set oldAvailableVarsError = new HashSet(this.availableVarsError);
        
        final String id = helper.getAttrValueNoEmpty(ruleElem, ClxConstants.ATTR_ID);
        boolean disabled = Boolean.valueOf(
                helper.getAttrValue(ruleElem, ClxConstants.ATTR_DISABLED, "false")).booleanValue();
        
        Element reportElement = null;
        ClxHeader header = null;
        ClxExtErrorReport report = null;
        ClxFormula formula = null;
        final String defaultErrMsg = "Rule " + id + " violated";
                
        List children = getChildrenFormulas(ruleElem);
        for(Iterator i = children.iterator(); i.hasNext(); ) {
            final Element child = (Element) i.next();
            if (!i.hasNext()) { //last element
                formula = createFormula(child);
            } else if (helper.isNS(child.getNamespaceURI())) {
                final String name = helper.getNodeName(child);
                if (ClxConstants.TAG_HEADER.equals(name)) {
                    header = createHeaderNull(child);
                } else if (ClxConstants.TAG_REPORT.equals(name)) {
                    reportElement = child;
                } else {
                    throw new ClxBuilderException("Unknown element in rule " + id, getErrorPath(child));
                }
            } else {
                throw new ClxBuilderException("Unknown element in rule " + id + 
                        " - invalid namespace", child.getNodeName());
            }
        }
        children = null;
        
        if (formula == null) {
            throw new ClxBuilderException("Formula definition is missing for rule " + id);
        }        
        if (reportElement != null) {
            report = createReport(reportElement, defaultErrMsg);
        } else {
            report = new ClxExtErrorReport(defaultErrMsg);
        }
                        
        this.availableVarsError = oldAvailableVarsError;
        
        return new ClxRule(id, disabled, header, report, formula);
    }
        
    /**
     * Creates a general header.
     * @param headerElem source - W3C DOM element - {@link ClxConstants#TAG_HEADER}, can be null 
     * @return header or null if source is null or {@link #skipHeader} is set to <code>true</code>
     */
    private ClxHeader createHeaderNull(Element headerElem) {
        if (skipHeader || (headerElem == null)) {
            return null;
        }
        
        final List authors = concatElemsValue(helper.getChildren(headerElem, ClxConstants.TAG_AUTHOR));        
        final List comments = concatElemsValue(helper.getChildren(headerElem, ClxConstants.TAG_COMMENT));
        
        List descElems = helper.getChildren(headerElem, ClxConstants.TAG_DESCRIPTION);
        final List descriptions = new ArrayList(descElems.size());
        for(Iterator i = descElems.iterator(); i.hasNext(); ) {
            final Element desc = (Element) i.next();
            descriptions.add(UtilDOM.formatWhitespaces(UtilDOM.printElementContent(desc), true, true));
        }
        
        return new ClxHeader(authors, descriptions, comments);
    }
              
    /**
     * Creates an error reporter.
     * @param reportElem source - W3C DOM element - {@link ClxConstants#TAG_REPORT}
     * @param defaultMessage default message returned in case when source doesn't define any
     *        message 
     * @return error reported; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxExtErrorReport createReport(Element reportElem, String defaultMessage)
    throws ClxBuilderException {
        List children = UtilDOM.getChildrenNS(reportElem, ClxConstants.TAG_EXT_ERRORREPORT, 
                ClxConstants.NS_EXT_REPORT, reportNoNS);
        for(Iterator i = children.iterator(); i.hasNext(); ) {
            final Element e = (Element) i.next();
            final String ns = e.getNamespaceURI();
            
            if (ClxConstants.NS_EXT_REPORT.equals(ns)) {
                return createExtErrorReport(e, ClxDOMHelperNS.newInstance(ns), defaultMessage);
            } else {
                return createExtErrorReport(e, ClxDOMHelper.newInstance(), defaultMessage);
            }
        }
        children = null;
                
        return new ClxExtErrorReport(UtilDOM.printElementContent(reportElem));
    }
    
    /**
     * Creates a message for extended error reporter.
     * @param msgElem source - W3C DOM element - {@link ClxConstants#TAG_EXT_ERRORMESSAGE}
     * @param repHelper to operate on {@link ClxConstants#NS_EXT_REPORT}
     * @return list of {@link ClxExtErrorMessage}; !=null. List is empty if source doesn't define
     *         any message 
     * @throws ClxBuilderException error in source
     */
    private List createExtErrorMessage(Element msgElem, ClxDOMHelper repHelper) 
    throws ClxBuilderException {
        NodeList nl = msgElem.getChildNodes();
        final int size = nl.getLength();
        final List result = new ArrayList(size);        
        
        for(int i = 0; i < size; i++) {
            final Node node = nl.item(i);
            
            switch(node.getNodeType()) {
            case Node.TEXT_NODE:
            case Node.CDATA_SECTION_NODE:
                final String msg = 
                    UtilDOM.formatWhitespaces(node.getNodeValue(), result.isEmpty(), i + 1 >= size);  
                if (msg.length() > 0) {
                    result.add(new ClxExtErrorMessageText(msg));
                }
                break;
                
            case Node.ELEMENT_NODE:
                final Element elem = (Element) node;
                final String name = repHelper.getNodeName(elem);
                if (repHelper.isNS(elem.getNamespaceURI())) {
                    if (ClxConstants.TAG_EXT_VALUEOF.equals(name)) {
                        String xpath = repHelper.getAttrValueNoEmpty(elem, ClxConstants.ATTR_XPATH);
                        String error = repHelper.getAttrValue(elem, ClxConstants.ATTR_ERROR, null);
                        result.add(new ClxExtErrorMessageXPath(
                                        createXPath(xpath, elem, this.availableVarsError), error));
                    } else {
                        throw new ClxBuilderException("Unknown element in error message", getErrorPath(elem));
                    }
                } else {
                    result.add(new ClxExtErrorMessageText(UtilDOM.printElement(elem)));
                }
                
            default:
                //skip all other elements
            }
        }
        
        return result;
    }
    
    /**
     * Creates an extended error reporter.
     * @param extReportElem source - W3C DOM element - {@link ClxConstants#TAG_EXT_ERRORREPORT}
     * @param repHelper to operate on {@link ClxConstants#NS_EXT_REPORT}
     * @param defaultMessage default message returned in case when source doesn't define any
     *        message 
     * @return extended error reported; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxExtErrorReport createExtErrorReport(Element extReportElem, ClxDOMHelper repHelper, 
            String defaultMessage)
    throws ClxBuilderException {
        //errorLevel
        int errorLevel = XdvValidationError.LEVEL_ERROR;
        Element levelElem = repHelper.getFirstOne(extReportElem, ClxConstants.TAG_EXT_ERRORLEVEL);
        if (levelElem != null) {
            final String value = repHelper.getAttrValueNoEmpty(levelElem, ClxConstants.ATTR_NAME);
            if (ClxConstants.VAL_ERRORLEVEL_ERROR.equals(value)) {
                errorLevel = XdvValidationError.LEVEL_ERROR;
            } else if (ClxConstants.VAL_ERRORLEVEL_WARNING.equals(value)) {
                errorLevel = XdvValidationError.LEVEL_WARNING;
            } else {
                throw new ClxBuilderException(value + " is not a valid error-level name", 
                        getErrorPath(levelElem));
            }            
        }
        levelElem = null; //I don't need this object any more        
        
        //errorNode
        List nodeElems = repHelper.getChildren(extReportElem, ClxConstants.TAG_EXT_ERRORNODE);
        final List errorNodes = new ArrayList(nodeElems.size());
        for(Iterator i = nodeElems.iterator(); i.hasNext(); ) {
            final Element elem = (Element) i.next();
            final String xpath = repHelper.getAttrValueNoEmpty(elem, ClxConstants.ATTR_XPATH);
            
            errorNodes.add(createXPath(xpath, elem, this.availableVarsError));
        }
        nodeElems = null; //I don't need this object any more
        
        //errorMessage
        List messages;
        Element msgElem = repHelper.getFirstOne(extReportElem, ClxConstants.TAG_EXT_ERRORMESSAGE);
        if (msgElem == null) {
            messages = new ArrayList(1); 
        } else {           
            messages = createExtErrorMessage(msgElem, repHelper);
        }
        msgElem = null; //I don't need this object any more
        
        if (messages.isEmpty()) {
            messages.add(new ClxExtErrorMessageText(defaultMessage));
        }
                        
        return new ClxExtErrorReport(errorLevel, messages, errorNodes);
    }
    
    /**
     * Creates a macro call and adds created object to {@link #createdMacroCalls}.
     * It creates an instance of {@link ClxExtMacroCallPostPrep} - which requires 
     * <i>post processing</i> before executing.
     * @param macroCallElem source - W3C DOM element - {@link ClxConstants#TAG_EXT_MACROCALL}
     * @param macroHelper to operate on {@link ClxConstants#NS_EXT_MACRO}
     * @return macro call; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxExtMacroCall createExtMacroCall(Element macroCallElem, ClxDOMHelper macroHelper) 
    throws ClxBuilderException {
        final String name = macroHelper.getAttrValueNoEmpty(macroCallElem, ClxConstants.ATTR_NAME);                 
        final Map params = new HashMap();
        
        //compute and compile parameters
        List paramElems = macroHelper.getChildren(macroCallElem, ClxConstants.TAG_PARAM);
        for(Iterator i = paramElems.iterator(); i.hasNext(); ) {
            final Element param = (Element) i.next();
            final String paramName = macroHelper.getAttrValueNoEmpty(param, ClxConstants.ATTR_NAME);            
            final XPathExpression xpath = createXPath(
                    macroHelper.getAttrValueNoEmpty(param, ClxConstants.ATTR_VALUE), param);
            
            Object oldVal = params.put(paramName, xpath);
            if (oldVal != null) {
                throw new ClxBuilderException("Parameter " + paramName + " defined more then once " +
                        "for macro " + name, getErrorPath(param));
            }
        }        
        paramElems = null; //I don't need this

        ClxExtMacroCallPostPrep macroCall = new ClxExtMacroCallPostPrep(name, params);
        this.createdMacroCalls.add(macroCall);
        return macroCall;
    }    

    /**
     * Creates a formula - it could be either a standard CLX formula or an extended formula
     * like a macroCall.
     * @param formulaElem source that describes a formula
     * @return formula; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxFormula createFormula(Element formulaElem) 
    throws ClxBuilderException {
        final String name = helper.getNodeName(formulaElem);
        
        //Checking namespace URI is useless - I rely only on local name
        if (ClxConstants.TAG_FORALL.equals(name)) {
            return createForAll(formulaElem);
        } else if (ClxConstants.TAG_EXISTS.equals(name)) {
            return createExists(formulaElem);
        } else if (ClxConstants.TAG_EQUAL.equals(name)) {
            return createEqual(formulaElem);
        } else if (ClxConstants.TAG_NOTEQUAL.equals(name)) {
            return createNotEqual(formulaElem);
        } else if (ClxConstants.TAG_SAME.equals(name)) {
            return createSame(formulaElem);
        } else if (ClxConstants.TAG_LESS.equals(name)) {
            return createLess(formulaElem);
        } else if (ClxConstants.TAG_GREATER.equals(name)) {
            return createGreater(formulaElem);
        } else if (ClxConstants.TAG_LESSOREQUAL.equals(name)) {
            return createLessOrEqual(formulaElem);
        } else if (ClxConstants.TAG_GREATEROREQUAL.equals(name)) {
            return createGreaterOrEqual(formulaElem);
        } else if (ClxConstants.TAG_AND.equals(name)) {
            return createAnd(formulaElem);
        } else if (ClxConstants.TAG_OR.equals(name)) {
            return createOr(formulaElem);
        } else if (ClxConstants.TAG_IMPLIES.equals(name)) {
            return createImplies(formulaElem);
        } else if (ClxConstants.TAG_IFF.equals(name)) {
            return createIff(formulaElem);
        } else if (ClxConstants.TAG_NOT.equals(name)) {
            return createNot(formulaElem);
        } else if (ClxConstants.TAG_OPERATOR.equals(name)) {
            return createOperator(formulaElem);
        } else if (ClxConstants.TAG_EXT_MACROCALL.equals(name)) {
            ClxDOMHelper macroHelper;
            if (ClxConstants.NS_EXT_MACRO.equals(formulaElem.getNamespaceURI())) {
                macroHelper = ClxDOMHelperNS.newInstance(ClxConstants.NS_EXT_MACRO);
            } else {
                macroHelper = ClxDOMHelper.newInstance();
            }
            
            return createExtMacroCall(formulaElem, macroHelper);
        }
        
        throw new ClxBuilderException("Unknown formula " + name, getErrorPath(formulaElem));
    }
    
    /**
     * Creates a formula - <b>forAll</b>, first order logic predicate.  
     * @param forAllElem source
     * @return formula; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxForAll createForAll(Element forAllElem) 
    throws ClxBuilderException {
        final String var = helper.getAttrValueNoEmpty(forAllElem, ClxConstants.ATTR_VAR);
        final String in = helper.getAttrValueNoEmpty(forAllElem, ClxConstants.ATTR_IN);
        
        declareVariable(var, forAllElem);
        
        //find and create formula
        List children = getChildrenFormulas(forAllElem);
        int size = children.size();
        ClxFormula formula;
        if (size == 0) {
            formula = new ClxFormulaTrue();
        } else if (size == 1) {
            formula = createFormula((Element) children.get(0));
        } else {
            throw new ClxBuilderException("Exacly one formula expected", getErrorPath(forAllElem));
        }        
        
        this.availableVars.remove(var);
        
        return new ClxForAll(var, createXPath(in, forAllElem), formula);
    }

    /**
     * Creates a formula - <b>exists</b>, first order logic predicate.  
     * @param existsElem source
     * @return formula; !=null
     * @throws ClxBuilderException error in source
     */
    private ClxExists createExists(Element existsElem) 
    throws ClxBuilderException {
        final String var = helper.getAttrValueNoEmpty(existsElem, ClxConstants.ATTR_VAR);
        final String in = helper.getAttrValueNoEmpty(existsElem, ClxConstants.ATTR_IN);
        
        declareVariable(var, existsElem);
        
        //find and compile formula
        List children = getChildrenFormulas(existsElem);
        int size = children.size();
        ClxFormula formula;
        if (size == 0) {
            formula = new ClxFormulaTrue();
        } else if (size == 1) {
            formula = createFormula((Element) children.get(0));
        } else {
            throw new ClxBuilderException("at most one child expected", getErrorPath(existsElem));
        }
        
        this.availableVars.remove(var);
        
        return new ClxExists(var, createXPath(in, existsElem), formula);
    }
    
    /**
     * Gets two parameters for binary predicate and creates tow XPath expressions from them.  
     * @param elem binary predicate definition
     * @return two XPath expressions 
     * @throws ClxBuilderException error in source
     */
    private BinaryPredicateParams getBinaryPredicateParams(Element elem) 
    throws ClxBuilderException {
        final BinaryPredicateParams result = new BinaryPredicateParams();
        result.op1 = createXPath(helper.getAttrValueNoEmpty(elem, ClxConstants.ATTR_OP1), elem);
        result.op2 = createXPath(helper.getAttrValueNoEmpty(elem, ClxConstants.ATTR_OP2), elem);
        return result;
    }

    /**
     * Creates an <b>equal</b> predicate.  
     * @param equalElem source - {@link ClxConstants#TAG_EQUAL}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxEqual createEqual(Element equalElem) 
    throws ClxBuilderException {
        BinaryPredicateParams params = getBinaryPredicateParams(equalElem);
        return new ClxEqual(params.op1, params.op2);
    }
   
    /**
     * Creates a <b>greather</b> predicate.  
     * @param greaterElem source - {@link ClxConstants#TAG_GREATER}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxGreater createGreater(Element greaterElem) 
    throws ClxBuilderException {
        BinaryPredicateParams params = getBinaryPredicateParams(greaterElem);
        return new ClxGreater(params.op1, params.op2);
    }
    
    /**
     * Creates a <b>greather or equal</b> predicate.  
     * @param greaterOrEqualElem source - {@link ClxConstants#TAG_GREATEROREQUAL}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxGreaterOrEqual createGreaterOrEqual(Element greaterOrEqualElem) 
    throws ClxBuilderException {
        BinaryPredicateParams params = getBinaryPredicateParams(greaterOrEqualElem);
        return new ClxGreaterOrEqual(params.op1, params.op2);
    }

    /**
     * Creates a <b>less</b> predicate.  
     * @param lessElem source - {@link ClxConstants#TAG_LESS}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxLess createLess(Element lessElem) 
    throws ClxBuilderException {
        BinaryPredicateParams params = getBinaryPredicateParams(lessElem);
        return new ClxLess(params.op1, params.op2);
    }
    
    /**
     * Creates a <b>less or equal</b> predicate.  
     * @param lessOrEqualElem source - {@link ClxConstants#TAG_LESSOREQUAL}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxLessOrEqual createLessOrEqual(Element lessOrEqualElem) 
    throws ClxBuilderException {
        BinaryPredicateParams params = getBinaryPredicateParams(lessOrEqualElem);
        return new ClxLessOrEqual(params.op1, params.op2);
    }
    
    /**
     * Creates a <b>not equal</b> predicate.  
     * @param notEqualElem source - {@link ClxConstants#TAG_NOTEQUAL}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxNotEqual createNotEqual(Element notEqualElem) 
    throws ClxBuilderException {
        BinaryPredicateParams params = getBinaryPredicateParams(notEqualElem);
        return new ClxNotEqual(params.op1, params.op2);
    }
    
    /**
     * Creates a <b>same</b> predicate.  
     * @param sameElem source - {@link ClxConstants#TAG_SAME}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxSame createSame(Element sameElem) 
    throws ClxBuilderException {
        BinaryPredicateParams params = getBinaryPredicateParams(sameElem);
        return new ClxSame(params.op1, params.op2);
    }
    
    /**
     * Creates two sub formulas for binary operator.  
     * @param elem binary operator definition, it has to have exacly 2 sub formulas
     * @return two sub formulas
     * @throws ClxBuilderException error in source
     */
    private BinaryOperatorParams getBinaryOperatorParams(Element elem) 
    throws ClxBuilderException {
        final BinaryOperatorParams params = new BinaryOperatorParams();
        
        //find two formulas
        List formulas = getChildrenFormulas(elem);
        if (formulas.size() != 2) {
            throw new ClxBuilderException("Exacly two children required", getErrorPath(elem));
        }        
        params.formula1 = createFormula((Element) formulas.get(0));
        params.formula2 = createFormula((Element) formulas.get(1));
        
        return params;
    }
    
    /**
     * Creates an <b>and</b> predicate.  
     * @param andElem source - {@link ClxConstants#TAG_AND}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxAnd createAnd(Element andElem) 
    throws ClxBuilderException {
        BinaryOperatorParams param = getBinaryOperatorParams(andElem);
        return new ClxAnd(param.formula1, param.formula2);
    }
    
    /**
     * Creates a <b>iff</b> (if and only if) predicate.  
     * @param iffElem source - {@link ClxConstants#TAG_IFF}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxIff createIff(Element iffElem) 
    throws ClxBuilderException {
        BinaryOperatorParams param = getBinaryOperatorParams(iffElem);
        return new ClxIff(param.formula1, param.formula2);
    }
    
    /**
     * Creates an <b>implies</b> predicate.  
     * @param impliesElem source - {@link ClxConstants#TAG_IMPLIES}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxImplies createImplies(Element impliesElem) 
    throws ClxBuilderException {
        BinaryOperatorParams param = getBinaryOperatorParams(impliesElem);
        return new ClxImplies(param.formula1, param.formula2);
    }
    
    /**
     * Creates an <b>or</b> predicate.  
     * @param orElem source - {@link ClxConstants#TAG_OR}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxOr createOr(Element orElem) 
    throws ClxBuilderException {
        BinaryOperatorParams param = getBinaryOperatorParams(orElem);
        return new ClxOr(param.formula1, param.formula2);
    }
    
    /**
     * Creates a <b>not</b> predicate.  
     * @param notElem source - {@link ClxConstants#TAG_NOT}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxNot createNot(Element notElem)
    throws ClxBuilderException {
        List elems = getChildrenFormulas(notElem);
        if (elems.size() != 1) {
            throw new ClxBuilderException("Exacly one child required", getErrorPath(notElem));
        }
        
        return new ClxNot(createFormula((Element) elems.get(0)));
    }
    
    /**
     * Creates an operator call and adds created object to {@link #createdOperatorCalls}.
     * It creates an instance of {@link ClxOperatorCallPostPrep} - which requires 
     * <i>post processing</i> before executing.
     * @param operatorElem source - {@link ClxConstants#TAG_OPERATOR}
     * @return predicate; !=null 
     * @throws ClxBuilderException error in source
     */
    private ClxOperatorCall createOperator(Element operatorElem)
    throws ClxBuilderException {        
        final String name = helper.getAttrValueNoEmpty(operatorElem, ClxConstants.ATTR_NAME);                
        
        //get parameters
        List paramElems = helper.getChildren(operatorElem, ClxConstants.TAG_PARAM);
        final List params = new ArrayList(paramElems.size());
        for(Iterator i = paramElems.iterator(); i.hasNext(); ) {
            final Element paramElem = (Element) i.next();
            final String paramName = helper.getAttrValueNoEmpty(paramElem, ClxConstants.ATTR_NAME);
            final String paramValue = helper.getAttrValueNoEmpty(paramElem, ClxConstants.ATTR_VALUE);
            params.add(new ClxOperatorParam(paramName, paramValue));            
        }
        paramElems = null; //I don't need this elemen in further computations

        ClxOperatorCallPostPrep opCall = new ClxOperatorCallPostPrep(name, params);
        this.createdOperatorCalls.add(opCall);
        return opCall;
    }
    
}

/**
 * Internal class to hold two arguments for binary predicates as XPath expressions.
 * @author Tomasz Kokoszka
 */
class BinaryPredicateParams {
    /** First argument */
    public XPathExpression op1;
    
    /** Secound argument */
    public XPathExpression op2;
}

/**
 * Internal class to hold two sub formulas for binary operators.
 * @author Tomasz Kokoszka
 */
class BinaryOperatorParams {
    /** First sub formula */
    public ClxFormula formula1;
    
    /** Secound sub formula */
    public ClxFormula formula2;
}
