package org.xdv.clx.exec;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.xdv.clx.expr.ClxMacro;
import org.xdv.common.XdvValidationException;
import org.xdv.xpath.XPathEngine;
import org.xdv.xpath.XPathException;
import org.xdv.xpath.XPathExecutionUnit;
import org.xdv.xpath.XPathExpression;
import org.xdv.xpath.XPathValue;
import org.xdv.xpath.XPathValueNodeSet;

/**
 * CLX execution context for one XML document.
 * <p>
 * Purpos of this clase is to delive convenient methods useful during executing CLX rule.
 * It also stores data required to execute CLX rule - 
 * variables mapping, engine to execute XPath expressions, XML document being validated, 
 * macros binding (used only to evalue <i>lazy macro calls</i>). 
 * </p>
 *
 * @author Tomasz Kokoszka
 */
public class ClxExecutionContext {
        
    /** XPath execution unit to execute xpath expressions */
    protected XPathExecutionUnit executionUnit;
    
    /** Macros binding - macro's name-&gt; macro's definition 
     * ({@link String}-&gt;{@link ClxMacro}). !=null */ 
    protected Map macros = new HashMap();    
    
    /** Set of all declared CLX variables (set of {@link String}) */
    protected Set declared = new HashSet();
       
    /** 
     * Create instance.
     * @param xpathEngine XPath engine to execute XPath expressions - it should be the same 
     *        engine instance that was used to create expressions
     * @param documentNative XML document being validated. Its class depends of type of
     *        XPath engine (each engine may accepts different type of XML document)
     */
    public ClxExecutionContext(XPathEngine xpathEngine, Object documentNative) {
        this.executionUnit = xpathEngine.createExecutionUnit(documentNative);
    }
    
    /**
     * Return XPath engine.
     * @return XPath engine; !=null
     */
    public XPathEngine getXPathEngine() {
        return this.executionUnit.getEngine();
    }
    
    /**
     * Sets completely new macros binding.
     * @param macros macros binding: macro's name-&gt; macro's definition 
     *        ({@link String}-&gt;{@link ClxMacro})
     */
    public void setMacros(Map macros) {
        this.macros = macros;
    }      
    
    /**
     * Get macro's definition.
     * @param name macro's name
     * @return macro's definition or <code>null</code> if there is no definition for given name
     */
    public ClxMacro getMacro(String name) {
        return (ClxMacro) macros.get(name);
    }
    
    /**
     * Creates and sets completely new CLX variable context.
     * @return previous CLX variable context; !=null
     */
    public ClxVariableContext setNewVariableContext() {
        ClxVariableContext old = 
            new ClxVariableContext(this.executionUnit.getVariables(), this.declared);        
        this.executionUnit.setVariables(this.executionUnit.getEngine().createVariableMap());
        this.declared = new HashSet();
        return old;
    }
    
    /**
     * Sets new CLX variable context.
     * @param newVarContext CLX variable context
     * @return previous CLX variable context; !=null
     */
    public ClxVariableContext setNewVariableContext(ClxVariableContext newVarContext) {
        ClxVariableContext old = 
            new ClxVariableContext(this.executionUnit.getVariables(), this.declared);        
        this.executionUnit.setVariables(newVarContext.getVariables());        
        this.declared = newVarContext.getDeclared();
        return old;
    }
     
    /**
     * Declare CLX variable.
     * @param name variable name
     * @throws XdvValidationException variable already declared
     */
    public void declareVariable(String name) 
    throws XdvValidationException {
        if (declared.contains(name)) {
            throw new XdvValidationException("Variable " + name + " already declared");
        }
        declared.add(name);
    }
        
    /**
     * Set CLX variable value.
     * @param name variable's name
     * @param value variable's value
     */
    public void setVariable(String name, XPathValue value) {
        this.executionUnit.getVariables().setVariable(name, value);        
    }        
    
    /**
     * Undeclare CLX variable.
     * Undeclared variables are not removed from variable binding. It is keeps in binding 
     * so it can be used to rich generate error message.
     * <br>To remove them call {@link #removeUndeclared()}.
     * @param name variable's name
     * @throws XdvValidationException variable has not been delcared before
     */
    public void undeclareVariable(String name) 
    throws XdvValidationException {
        if (! declared.remove(name)) {
            throw new XdvValidationException("Variable " + name + " is not declared");
        }
    }    
    
    /**
     * Cleans variable context - removes all undeclared variable binding.
     */
    public void removeUndeclared() {
        //I need a copy because otherwise I may get ConcurrentModificationException 
        // - depends on XPath implementation
        final Collection names = new ArrayList(this.executionUnit.getVariables().getNames());
        for(Iterator i = names.iterator(); i.hasNext(); ) {
            final String name = (String) i.next();
            if (! declared.contains(name)) {
                this.executionUnit.getVariables().removeVariable(name);
            }
        }
    }
            
    /** 
     * Execute XPath expression.
     * @param xpathExpression expression to execute
     * @return execution result; !=null
     * @throws XPathException error while executing expression
     */
    public XPathValue executeXPath(XPathExpression xpathExpression) 
    throws XPathException {
        return executionUnit.execute(xpathExpression);
    }
    
    /**
     * Execute XPath expression and get result as a node-set.
     * @param xpathExpression expression to execute. It should evaluate to a node-set, otherwise
     *        a {@link XPathException} will be thrown
     * @return execution result; !=null 
     * @throws XPathException error while executing expressions
     */
    public XPathValueNodeSet executeXPathSelect(XPathExpression xpathExpression)
    throws XPathException {
        return executionUnit.executeSelect(xpathExpression);
    }
        
}
