package org.xdv.clx.expr;

import java.util.ListIterator;

import org.xdv.clx.exec.ClxExecutionContext;
import org.xdv.clx.exec.ClxExecutionPoint;
import org.xdv.common.XdvValidationException;
import org.xdv.xpath.XPathException;
import org.xdv.xpath.XPathExpression;
import org.xdv.xpath.XPathValueNodeSet;

/**
 * Quantifier - a general CLX type.
 * It consist of variable's name, XPath expression that is evaluated to a node-set, one sub-formula.
 *
 * @see <a href="http://www.clixml.org/clix/1.0/docs/html/#complexType_QuantifierType_Link03693610">CliXML QuantifierType</a> 
 * @author Tomasz Kokoszka
 */

public abstract class ClxFormulaQuantifier implements ClxFormula {

    /** Variable name; != null */
    private final String var;
    
    /** XPath expression that is evaluated to a node-set; !=null */
    private final XPathExpression in;
    
    /** Sub-formula; !=null */
    private final ClxFormula formula;

    /**
     * Creates instance.
     * @param var variable's name
     * @param in XPath expression that is evaluated to a node-set
     * @param formula sub-formula
     */
    public ClxFormulaQuantifier(String var, XPathExpression in, ClxFormula formula) {
        super();
        this.var = var;
        this.in = in;
        this.formula = formula;
    }
    
    /**
     * Returns qunatifier name.
     * @return quantifier name; !=null
     */
    public abstract String getName();

    /**
     * Returns XPath expression that is evaluated to a node-set
     * @return XPath expression; !=null
     */
    public XPathExpression getIn() {
        return in;
    }

    /**
     * Returns variable's name.
     * @return variable's name; !=null
     */
    public String getVar() {
        return var;
    }        
    
    /**
     * Returns sub-formula.
     * @return sub-formula; !=null
     */
    public ClxFormula getFormula() {
        return this.formula;
    }
    
    /**
     * Returns execution state from given execution point, if execution state doesn't exist
     * it creates a new one (by executing {@link #in}) and puts it into execution point.
     * @param cntx CLX execution context
     * @param frame execution point
     * @return execution state - ready to use iterator through all nodes that this quantifier
     *         checks - iterator of {@link XPathValueNodeSet}
     * @throws XdvValidationException error while executing {@link #in}
     */
    protected ListIterator getExecutionState(ClxExecutionContext cntx, ClxExecutionPoint frame)
    throws XdvValidationException {
        ListIterator iter = (ListIterator) frame.getData();
        if (iter == null) {
            XPathValueNodeSet nodes = getInNodeSet(cntx);
            iter = nodes.getNodesAsValues().listIterator();
            frame.setData(iter);                                    
        }
        return iter;
    }

    /**
     * Evaluates {@link #in} xpath and returns result as a node-set.
     * @param cntx CLX execution context
     * @return node-set; !=null
     * @throws XdvValidationException error while executing {@link #in}
     */
    protected XPathValueNodeSet getInNodeSet(ClxExecutionContext cntx)
    throws XdvValidationException {
        try {
            return cntx.executeXPathSelect(getIn());
        } catch (XPathException e) {
            throw new XdvValidationException("Error while executing xpath for quantifier " +
                    "(" + getName() + " $" + getVar() + " in " + getIn().asString() + ")", e);
        }        
    }
    
    public String toString() {
        return getName() + " $" + getVar() + " in " + getIn().asString() + " (" + getFormula() + ")";
    }
    
}
