package org.xdv.xpath;

import java.util.Set;

/**
 * XPath variable map - mapping between variable name and variable value.
 * <p>
 * This map is used to resolve variables in XPath expression.
 * For instance: expression <code>/data/field[substring(@name, 1, 2)=$A]</code> requires
 * variable <code>A</code> which have to be defined before one will try to execute this expression.
 * </p>
 * <p>
 * To work with namespaces use methods where you can explicity specify the namespace URI. 
 * It is because there is no URI for <i>default</i> namespace.
 * </p>
 *
 * @author Tomasz Kokoszka
 */
public interface XPathVariableMap {
    
    /**
     * Sets value of variable from <i>default</i> namespace. 
     * If variable already exists in the map then its value is updated, otherwise variable is added
     * and set to given value.
     * @param varName variable name 
     * @param varValue variable value
     * @return Previous variable value or <code>null</code> if variable was not in map before 
     */
    public XPathValue setVariable(String varName, XPathValue varValue);

    /**
     * Returns value of variable from <i>default</i> namespace.  
     * @param varName variable name
     * @return Variable value or <code>null</code> if variable is not in map
     */
    public XPathValue getVariable(String varName);
    
    /**
     * Removes variable from <i>default</i> namespace.
     * If variable is not in the map nothing happens.
     * @param varName variable name
     * @return Value of removed variable or <code>null</code> if variable is not in the map
     */
    public XPathValue removeVariable(String varName);
    
    /**
     * Returns names of all variables from <i>default</i> namespace.
     * @return Set of {@link String}. !=null. If <i>default</i> namespace is empty then an empty 
     *         set is returned.
     */
    public Set getNames();

    /**
     * Sets new variable value. 
     * If variable already exists in map then its value is updated, otherwise variable is added
     * and set to given value.
     * @param varName variable name
     * @param namespaceURI variable namespace 
     *        (for instance: <code>http://www.w3c.org/schema</code>) 
     * @param varValue variable value
     * @return Previous variable value or <code>null</code> if variable was not in the map before 
     */
    public XPathValue setVariable(String varName, String namespaceURI, XPathValue varValue);

    /**
     * Returns variable value.  
     * @param varName variable name
     * @param namespaceURI variable namespace 
     *        (for instance: <code>http://www.w3c.org/schema</code>) 
     * @return Variable value or <code>null</code> if variable is not in the map
     */
    public XPathValue getVariable(String varName, String namespaceURI);
    
    /**
     * Removes variable.
     * If variable was not in the map nothing happens.
     * @param varName variable name
     * @param namespaceURI variable namespace 
     *        (for instance: <code>http://www.w3c.org/schema</code>)
     * @return Value of removed variable or <code>null</code> if variable was not in the map
     */
    public XPathValue removeVariable(String varName, String namespaceURI);

    /**
     * Returns names of all variables from given namespace.
     * @param namespaceURI namespace URI (for instance: <code>http://www.w3c.org/schema</code>)
     * @return set of {@link String}. !=null. If namespace is empty then an empty set is returned.
     */
    public Set getNames(String namespaceURI);
        
}
