package org.xdv.xpath.common;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.xdv.xpath.XPathValue;
import org.xdv.xpath.XPathVariableMap;

/**
 * XPath variable map, mapping between variable name and variable value ({@link XPathValue}).
 * <p>
 * It stores two independed maps:
 * <ol class="NoTopMargin">
 * <li>Map for <i>default</i> namespace: variable name-&gt;variable value
 * <li>Map for all other namespaces: (variable name, variable namespace URI) -&gt; variable value.
 * </ol>
 * Those two maps are completely independed in a way that variable from 
 * <i>default</i> namespace cannot be accesed using methods with namespace URI and vice versa.
 * It is a consequence of the fact that <i>default</i> namespace doesn't have URI.
 * </p>
 *
 * @author Tomasz Kokoszka
 */
public class BaseXPathVariableMap implements XPathVariableMap {

    /**
     * Full qualified variable name - variable name + namespace URI.
     * @author Tomasz Kokoszka
     */
    public static class VarNameNS {
        
        /** Variable name */
        private final String localName;
        
        /** Variable namespace URI */
        private final String namespaceURI;

        /** 
         * Creates new full variable name.
         * @param localName variable name
         * @param namespaceURI variable name space URI
         */
        public VarNameNS(String localName, String namespaceURI) {
            this.namespaceURI = namespaceURI;
            this.localName = localName;
        }
        
        /**
         * Returns variable.
         * @return Variable name; !=null
         */
        public String getLocalName() {
            return localName;
        }

        /**
         * Variable namespace.
         * @return Namespace URI; !=null
         */
        public String getNamespaceURI() {
            return namespaceURI;
        }

        public int hashCode() {
            return (localName.hashCode() ^ namespaceURI.hashCode());
        }

        public boolean equals(Object o) {
            if (o instanceof VarNameNS) {
                VarNameNS other = (VarNameNS) o;
                return (namespaceURI.equals(other.namespaceURI) && other.localName.equals(localName));
            }
            return false;
        }
        
        /**
         * Returns full qualified variable name.
         * @return Uri + ':' + variable name
         */
        public String toString() {
            return namespaceURI + ":" + localName;
        }

    }

    /** Variables in <i>default</i> name space, map: {@link String}-&gt;{@link XPathValue}. */
    protected final Map variables = new HashMap();

    /** Variables outside <i>default</i> name space, map: {@link VarNameNS}-&gt;{@link XPathValue}. */
    protected final Map variablesNS = new HashMap();

    /**
     * Creates empty variable map.
     */
    public BaseXPathVariableMap() {
    }
    
    public XPathValue getVariable(String varName) {
        return (XPathValue) variables.get(varName);
    }

    public XPathValue removeVariable(String varName) {
        return (XPathValue) variables.remove(varName);
    }

    public XPathValue setVariable(String varName, XPathValue varValue) {
        return (XPathValue) variables.put(varName, varValue);
    }

    public XPathValue getVariable(String varName, String namespaceURI) {
        return (XPathValue) variablesNS.get(new VarNameNS(varName, namespaceURI));
    }

    public XPathValue removeVariable(String varName, String namespaceURI) {
        return (XPathValue) variablesNS.remove(new VarNameNS(varName, namespaceURI));
    }

    public XPathValue setVariable(String varName, String namespaceURI, XPathValue varValue) {
        return (XPathValue) variablesNS.put(new VarNameNS(varName, namespaceURI), varValue);
    }
    
    public Set getNames() {
        return variables.keySet();
    }
    
    public Set getNames(String namespaceURI) {
        final Set result = new HashSet();
        for(Iterator i = variablesNS.keySet().iterator(); i.hasNext(); ) {
            final VarNameNS varName = (VarNameNS) i.next();
            if (namespaceURI.equals(varName.getNamespaceURI())) {
                result.add(varName.getLocalName());
            }
        }
        
        return result;
    }

}
